-----------------------------------------------------------
-- WXFruit: an implementation of Fruit using wxHaskell.
--   Version 0.1, May 2004
--   Contributors: Antony Courtney, Bart Robinson
-----------------------------------------------------------

module WXFruit where

-----------------------------------------------------------
-- Module imports.

import qualified Graphics.UI.WX as WX
import qualified Graphics.UI.WXCore as WXCore
import AFRP
import AFRPUtilities
import Control.Arrow

import Control.Concurrent.Chan
import GHC.Exts
import Data.IORef
import Data.Map as Map
import Foreign.Marshal.Utils
import System.Time


-----------------------------------------------------------
-- Data types for communication between widgets and the
-- system.  (These are likely to change dramatically.)

-- Raw input: kind of a hack...
data RawInput = RawInput {
     mousePos :: WX.Point, -- relative to topmost frame
     windowSize :: WX.Size,
     rawEvent :: Event RawEvent
     -- add other raw inputs here: keyboard, mouse presses?
  } deriving Show

-- for debug print
instance Show a => Show (Event a) where
    show NoEvent = "NoEvent"
    showsPrec d (Event m) = showParen (d > app_prec) $
             showString "Event " . showsPrec (app_prec+1) m
          where app_prec = 10

-- Refer from Graphics.HGL.Window
data RawEvent = Key {
                  keySym :: WX.EventKey,
                  isDown :: Bool
              }
              | Mouse {
                  mouseEvent :: WX.EventMouse
              } deriving Show

-- Responses/Requests: This is rather clumsy, as any new
-- widgets would have to add their own alternatives to
-- these sums.  A way to do more dynamic run-time typing
-- could help with this problem.
data WidgetResp = Init -- currently not used
                | ButtonCreateResp (WX.Button ())
                | TextCreateResp (WX.StaticText ())
                | PictureCreateResp (WX.Panel ())
                | GLPictureCreateResp (WXCore.GLCanvas ())
                | SliderCreateResp (WX.Slider ())
                | ButtonCommandEvent -- button pressed
                | SliderCommandEvent Int -- new value selected
                -- Menu
                | MenuCreateResp (WX.MenuItem ())
                | MenuCommandEvent
                | DeComp (Event WidgetResp,Event WidgetResp)

data WidgetReq = ButtonCreateReq ButtonState
               | TextCreateReq TextState
               | PictureCreateReq PictureState
               | GLPictureCreateReq GLPictureState
               | SliderCreateReq SliderState
               | ButtonSetReq (WX.Button ()) ButtonState
               | TextSetReq (WX.StaticText ()) TextState
               | PictureSetReq (WX.Panel ()) PictureState
               | GLPictureSetReq (WXCore.GLCanvas ()) GLPictureState
               | SliderSetReq (WX.Slider ()) SliderState
               -- Menu
               | MenuCreateReq MenuState
               | MenuSetReq (WX.MenuItem ()) MenuState
               -- command
               -- | CommandSetReq CommandState
               | Comp Orientation (Event WidgetReq,Event WidgetReq)

data Orientation = Horiz | Vert
  deriving (Eq)


-----------------------------------------------------------
-- Widgets, GUIs, and Boxes.  The basic WXFruit building
-- blocks.

-- Widget: just a fancy signal function.
type SFWidget a b = SF (RawInput,Event WidgetResp,a) (Event WidgetReq,b)

-- WXGUI: conceptually the same as a Widget, but a newtype
-- is used so that it can be made into an Arrow.
newtype GUI b c = GUI (SFWidget b c)

-- WXGUI functions for Arrow instance.  This makes WXGUI
-- correspond to the GA type in the original Fruit.  I'm
-- not actually not sure this is necessary, because the
-- preferred idiom is to use WXBoxes, anyway.

-- Lifting/lowering

unGUI :: GUI b c -> SFWidget b c
unGUI (GUI w) = w

wxSF :: SF b c -> GUI b c
wxSF sf = GUI $ proc (_,_,b) -> do
  c <- sf -< b
  returnA -< (noEvent,c)

wxArr :: (b -> c) -> GUI b c
wxArr f = wxSF (arr f)

-- Composing: handles the "plumbing" involved in composing
-- widgets together -- makes sure the system input/output
-- streams are hooked up correctly.  The ugliness this
-- hides is basically the motivation for creating an Arrow
-- instance for WXGUI (and WXBox).

decompResp :: Event WidgetResp -> (Event WidgetResp,Event WidgetResp)
decompResp (Event Init) = (Event Init,Event Init)
decompResp (Event (DeComp (resp0,resp1))) = (resp0,resp1)
decompResp _ = (noEvent,noEvent)

compReq :: Orientation -> Event WidgetReq -> Event WidgetReq -> Event WidgetReq
compReq _ NoEvent NoEvent = noEvent
compReq orient req0 req1 = Event (Comp orient (req0,req1))

-- Composition generalized over orientations
gComp :: Orientation -> GUI b c -> GUI c d -> GUI b d
gComp orient (GUI w1) (GUI w2) = GUI $ proc (inp,resp,b) -> do
  let (resp1,resp2) = decompResp resp
  (req1,c) <- w1 -< (inp,resp1,b)
  (req2,d) <- w2 -< (inp,resp2,c)
  returnA -< (compReq orient req1 req2,d)

-- Default layout for composed widgets is horizontal.
wxComp :: GUI b c -> GUI c d -> GUI b d
wxComp = gComp Horiz

wxFirst :: GUI b c -> GUI (b,d) (c,d)
wxFirst (GUI w) = GUI $ proc (inp,resp,(b,d)) -> do
  (req,c) <- w -< (inp,resp,b)
  returnA -< (req,(c,d))

instance Arrow GUI where
  arr = wxArr
  (>>>) = wxComp
  first = wxFirst

wxLoop :: GUI (b,d) (c,d) -> GUI b c
wxLoop (GUI w) = GUI $ proc (inp,resp,b) -> do
  rec (req,(c,d)) <- w -< (inp,resp,(b,d))
  returnA -< (req,c)

instance ArrowLoop GUI where
  loop = wxLoop

-- WXBox: Builds upon WXGUI, adding the ability to defer
-- the specification of the orientation of a collection of
-- widgets.

newtype Box b c = Box (Orientation -> GUI b c)

-- Lift a GUI into a Box, keeping the GUI's internal
-- orientation(s)
box :: GUI b c -> Box b c
box w = Box (const w)

-- Compose Boxes by threading orientation to children.
compBox :: Box b c -> Box c d -> Box b d
compBox (Box b1f) (Box b2f) = Box $
  \orient -> gComp orient (b1f orient) (b2f orient)

firstBox :: Box b c -> Box (b,d) (c,d)
firstBox (Box bf) = Box $ \orient -> first (bf orient)

instance Arrow Box where
  arr = box . arr
  (>>>) = compBox
  first = firstBox

wxLoopBox :: Box (b,d) (c,d) -> Box b c
wxLoopBox (Box bf) = Box $ \orient -> loop (bf orient)

instance ArrowLoop Box where
  loop = wxLoopBox

-- Some useful Box functions

boxGUI :: SFWidget b c -> Box b c
boxGUI = box . GUI

boxSF :: SF b c -> Box b c
boxSF = box . wxSF

-- These two are used to surround collections of Boxes and
-- make them go left-to-right or top-to-bottom.  They both
-- have the same effect when applied to a WXBox containing
-- just a single widget.

hBox :: Box b c -> GUI b c
hBox (Box bf) = bf Horiz

vBox :: Box b c -> GUI b c
vBox (Box bf) = bf Vert


-----------------------------------------------------------
-- Individual Widget definitions.  These are defined as
-- WXBoxes to allow easy composition.  They all employ a
-- similar pattern: they have a state-modifier function as
-- input, and use the same internal idioms to communicate
-- with the system.  See Button for extensive comments.
--   As more of these are added, it might be helpful to
-- move them into other modules/files.  (Which might be
-- tricky, as they are currently closed tied up with the
-- response/request data types.)

-- Utility to detect when a widget's state has changed
maybeChanged :: Eq a => a -> a -> Maybe ()
maybeChanged s s' = if s == s' then Nothing else Just ()

eqFunction :: a -> a -> Bool
eqFunction a b = toBool $ I# (reallyUnsafePtrEquality# a b)

dialogOrNothing :: a -> (a -> Bool) -> (a -> IO ()) -> (a -> WX.Frame () -> IO ()) -> (WX.Frame ()) -> IO ()
dialogOrNothing state e elsecom thencom f =
   if (e state)
     then (thencom state) f
     else (elsecom state)


-----------------------------------------------------------
-- Button widget: has a string label, can be enabled and
-- disabled, and outputs an Event () when pressed

-- State

data ButtonState = ButtonState {
     bsLabel :: String,
     bsEnabled :: Bool,
     bsAddedCommand :: IO (),
     bsAddedCommandWithDialog :: WX.Frame () -> IO (),
     bsEnableDialog :: Bool
  } -- deriving (Eq, Show)

instance Eq ButtonState where
    (==) (ButtonState ll lb lc lcd led)
         (ButtonState rl rb rc rcd red)
      = ll == rl && lb == rb
      && eqFunction lc rc && eqFunction lcd rcd && led == red

type ButtonConf = ButtonState -> ButtonState

-- Constructors to initialize and alter state.  Note that
-- the state of a button should not (cannot?) be modified
-- other than through the use of these functions.

btext :: String -> ButtonConf
btext l bs = bs {bsLabel = l}

btextWithDialog :: String -> ButtonConf
btextWithDialog l bs = bs {bsLabel = l, bsEnableDialog = True}

benabled :: Bool -> ButtonConf
benabled e bs = bs {bsEnabled = e}

bcommand :: IO () -> ButtonConf
bcommand c bs = bs {bsAddedCommand = c}

bcommandWithDialog  :: (WX.Frame () -> IO ()) -> ButtonConf
bcommandWithDialog  c bs = bs {bsAddedCommandWithDialog = c}

bcommandText :: IO () -> String -> ButtonConf
bcommandText c t bs = bs {bsAddedCommand = c, bsLabel = t}

bcommandTextWithDialog :: (WX.Frame () -> IO ()) -> String -> ButtonConf
bcommandTextWithDialog c t bs = bs {bsAddedCommandWithDialog = c, bsLabel = t}

benableDialog :: Bool -> ButtonConf
benableDialog e bs = bs {bsEnableDialog = e}

-- The button GUI

button :: ButtonConf -> Box ButtonConf (Event ())
button conf0 =
  let -- Initial state
      defState = ButtonState { bsLabel = "Default", bsEnabled = True, bsEnableDialog = False
                             , bsAddedCommand = return (), bsAddedCommandWithDialog = \f -> return ()}
      initState = conf0 defState
      
      -- Detect creation
      maybeCreate (ButtonCreateResp b) = Just (True,b)
      maybeCreate _ = Nothing
      
      -- Detect button press
      maybeCommand ButtonCommandEvent = Just ()
      maybeCommand _ = Nothing
  
  -- Make an ordinary, atomic signal function (in order to have access
  -- to the response/request mechanism), then lift it into a WXBox
  -- with wxBoxGUI.
  in boxGUI $ proc (_,resp,conf) -> do
    -- Keep track of the state.
    rec state <- iPre initState -< conf state
    
    -- Has the state changed?  If so, generate set request.
    stateChanged <- edgeBy maybeChanged initState -< state
    
    -- Ensure that the system hands us back a WX.Button ().
    (isCreated,button) <- hold (False,undefined) -< mapFilterE maybeCreate resp
        
    -- Send a creation request if we haven't been created yet.
    -- WARNING: this is probably a bad way to do this.  It can
    -- create an infinitely-dense stream of Events, which is a
    -- big no-no in the AFRP world (at least, conceptually).
    -- We used to use WXWInit for this purpose, but with the
    -- dynamic switching of widgets, the system wouldn't know
    -- when to send the WXWInit events to widgets that have
    -- been switched-into.
    let doCreate = if isCreated then noEvent else Event ()
    
    -- Merge create/set requests.
    let req = lMerge (tag doCreate (ButtonCreateReq state))
                     (tag stateChanged (ButtonSetReq button state))
    
    -- Pass button presses through.
    let press = mapFilterE maybeCommand resp
    
    returnA -< (req,press)


-----------------------------------------------------------
-- Static text (label) widget: Displays a string, and has
-- no output.

-- State

data TextState = TextState {            
     tsLabel :: String
  } deriving (Eq, Show)

type TextConf = TextState -> TextState

-- Constructor

ttext :: String -> TextConf
ttext l ts = ts {tsLabel = l}

-- The text GUI

text :: TextConf -> Box TextConf ()
text conf0 = 
  let -- Initial state
      defState = TextState {tsLabel = "Default"}
      initState = conf0 defState
      
      -- Detect creation
      maybeCreate (TextCreateResp t) = Just (True,t)
      maybeCreate _ = Nothing
  
  in boxGUI $ proc (_,resp,conf) -> do
    -- State
    rec state <- iPre initState -< conf state
    stateChanged <- edgeBy maybeChanged initState -< state
    
    -- Creation
    (isCreated,text) <- hold (False,undefined) -< mapFilterE maybeCreate resp
    let doCreate = if isCreated then noEvent else Event ()
    
    -- Output
    let req = lMerge (tag doCreate (TextCreateReq state))
                     (tag stateChanged (TextSetReq text state))
    returnA -< (req,())


-----------------------------------------------------------
-- Slider widget: a slider that the user can move along a
-- bar between minimum and maximum values.  Can be oriented
-- horizontally or vertically (upon creation), and
-- enabled/disabled dynamically.  Outputs a continuous
-- signal indicating the currently-selected value.

-- State

data SliderState = SliderState {
     -- These properties are determined upon creation,
     -- and cannot be subsequently altered.
     ssMin :: Int,
     ssMax :: Int,
     ssOrientation :: Orientation,
     
     ssSelection :: Int, -- user- and programmer-modifiable
     ssEnabled :: Bool -- programmer-modifiable
  } deriving (Eq)

type SliderConf = SliderState -> SliderState

-- Constructors

senabled :: Bool -> SliderConf
senabled e ss = ss {ssEnabled = e}

sselection :: Int -> SliderConf
sselection sel ss = ss {ssSelection = sel}

-- The slider GUI

slider :: Orientation -> Int -> Int -> SliderConf -> Box SliderConf Int
slider orient min max conf0 =
  let -- Initial state
      defState = SliderState
        {ssMin = min, ssMax = max, ssOrientation = orient,
         ssSelection = 0, ssEnabled = True}
      initState = conf0 defState
      
      -- Change state when user moves slider
      changeSel state (Event (SliderCommandEvent sel)) =
        state {ssSelection = sel}
      changeSel state _ = state
      
      -- Define our own maybeChanged, because we don't want to generate
      -- a WXWSliderSetReq if the only change in the state that occurred
      -- was via user input.
      maybeChangedInternal (_,state) (Event (SliderCommandEvent _), state') =
        if (state {ssSelection = ssSelection state'}) == state'
          then Nothing
          else Just () 
      maybeChangedInternal (_,state) (_,state') =
        maybeChanged state state'
      
      -- Detect creation
      maybeCreate (SliderCreateResp s) = Just (True,s)
      maybeCreate _ = Nothing
      
  in boxGUI $ proc (_,resp,conf) -> do
    -- State
    rec let newState = changeSel (conf state) resp
        state <- iPre initState -< newState
    
    -- Detect actionable state changes
    stateChanged <- edgeBy maybeChangedInternal (noEvent,initState) -< (resp,state)
    
    -- Creation
    (isCreated,slider) <- hold (False,undefined) -< mapFilterE maybeCreate resp
    let doCreate = if isCreated then noEvent else Event ()
    
    -- Output
    let req = lMerge (tag doCreate (SliderCreateReq state))
                     (tag stateChanged (SliderSetReq slider state))
    returnA -< (req,ssSelection state)


-----------------------------------------------------------
-- Picture widget: a box for doing arbitrary graphics.
-- Uses a WX.Panel () with a custom repaint handler.

-- State

data PictureState = PictureState {
    psSize :: WX.Size,
    psPicture :: Picture
  }

type PictureConf = PictureState -> PictureState

-- Constructors

ppic :: Picture -> PictureConf
ppic pic ps = ps {psPicture = pic}

psize :: WX.Size -> PictureConf
psize size ps = ps {psSize = size}

-- The picture GUI

picture :: PictureConf -> Box PictureConf ()
picture conf0 =
  let -- Initial state
      defState = PictureState {psSize = WX.sizeNull, psPicture = blank}
      initState = conf0 defState
      
      -- Detect creation
      maybeCreate (PictureCreateResp panel) = Just (True,panel)
      maybeCreate _ = Nothing
      
  in boxGUI $ proc (_,resp,conf) -> do
    -- State
    rec state <- iPre initState -< conf state
    
    -- Creation
    (isCreated,panel) <- hold (False,undefined) -< mapFilterE maybeCreate resp
    let doCreate = if isCreated then noEvent else Event ()
    
    -- Output.  Note that here we are again violating the
    -- rules of AFRP by sending an infinitely-dense stream
    -- of events, because we want a new picture at every
    -- sample time so that animation will work.
    let req = lMerge (tag doCreate (PictureCreateReq state))
                     (Event (PictureSetReq panel state))
    returnA -< (req,())


-----------------------------------------------------------
-- GLPicture widget: a box for doing arbitrary graphics.
-- Uses a WX.Panel () with a custom repaint handler.

-- State

data GLPictureState = GLPictureState {
    glpsSize :: WX.Size,
    glpsAttributes :: [WXCore.GLAttribute],
    glpsPicture :: GLPicture,
    glpsPictureWithDialog :: WX.Frame () -> GLPicture,
    glpsCanvas :: WXCore.Frame () -> WX.Size -> [WXCore.GLAttribute] -> IO (WXCore.GLCanvas ()),
    glpsEnableDialog :: Bool
  }

-- WXCore.GLAttribute doen't derive from Eq class.
instance Eq WXCore.GLAttribute where
    WXCore.GL_RGBA	             == WXCore.GL_RGBA               = True
    WXCore.GL_BUFFER_SIZE li     == WXCore.GL_BUFFER_SIZE ri     = li == ri
    WXCore.GL_LEVEL lo           == WXCore.GL_LEVEL ro           = lo == ro
    WXCore.GL_DOUBLEBUFFER       == WXCore.GL_DOUBLEBUFFER       = True
    WXCore.GL_STEREO             == WXCore.GL_STEREO             = True
    WXCore.GL_AUX_BUFFERS li     == WXCore.GL_AUX_BUFFERS ri     = li == ri
    WXCore.GL_MIN_RED li         == WXCore.GL_MIN_RED ri         = li == ri
    WXCore.GL_MIN_GREEN li       == WXCore.GL_MIN_GREEN ri       = li == ri
    WXCore.GL_MIN_BLUE li        == WXCore.GL_MIN_BLUE ri        = li == ri
    WXCore.GL_MIN_ALPHA li       == WXCore.GL_MIN_ALPHA ri       = li == ri
    WXCore.GL_DEPTH_SIZE li      == WXCore.GL_DEPTH_SIZE ri      = li == ri
    WXCore.GL_STENCIL_SIZE li    == WXCore.GL_STENCIL_SIZE ri    = li == ri
    WXCore.GL_MIN_ACCUM_RED li   == WXCore.GL_MIN_ACCUM_RED ri   = li == ri
    WXCore.GL_MIN_ACCUM_GREEN li == WXCore.GL_MIN_ACCUM_GREEN ri = li == ri
    WXCore.GL_MIN_ACCUM_BLUE li  == WXCore.GL_MIN_ACCUM_BLUE ri  = li == ri
    WXCore.GL_MIN_ACCUM_ALPHA li == WXCore.GL_MIN_ACCUM_ALPHA ri = li == ri
    _                            == _                            = False

instance Eq GLPictureState where
    (==) (GLPictureState ls la lp lpd lc le)
         (GLPictureState rs ra rp rpd rc re)
      = ls == rs && la == ra && eqFunction lp rp && eqFunction lpd rpd &&  eqFunction lc rc &&le == re

type GLPictureConf = GLPictureState -> GLPictureState

-- Constructors

glppic :: GLPicture -> GLPictureConf
glppic pic ps = ps {glpsPicture = pic}

glppicWithDialog :: (WX.Frame () -> GLPicture) -> GLPictureConf
glppicWithDialog pic ps = ps {glpsPictureWithDialog = pic, glpsEnableDialog = True}

glpsize :: WX.Size -> GLPictureConf
glpsize size ps = ps {glpsSize = size}

glpattributes :: [WXCore.GLAttribute] -> GLPictureConf
glpattributes attrs ps = ps {glpsAttributes = attrs}

glpsizeWithAttributes :: WX.Size -> [WXCore.GLAttribute] -> GLPictureConf
glpsizeWithAttributes size attrs ps = ps {glpsSize = size, glpsAttributes = attrs}

glpcanvas :: (WXCore.Frame () -> WX.Size -> [WXCore.GLAttribute] -> IO (WXCore.GLCanvas ())) -> GLPictureConf
glpcanvas gl ps = ps {glpsCanvas = gl}

glpenableDialog :: Bool -> GLPictureConf
glpenableDialog e ps = ps {glpsEnableDialog = e}

-- The picture GUI

glpicture :: GLPictureConf -> Box GLPictureConf ()
glpicture conf0 =
  let -- Initial state
      defState = GLPictureState { glpsSize = WX.sizeNull, glpsAttributes = [WXCore.GL_RGBA]
                                , glpsPicture = glBlank
                                , glpsPictureWithDialog = \f -> glBlank
                                , glpsCanvas = \f size attributes ->
                                    WXCore.glCanvasCreateEx f 0 (WX.Rect 0 0 (WX.sizeW size) (WX.sizeH size)) 0
                                    "GLCanvas" attributes WXCore.nullPalette
                                , glpsEnableDialog = False}
      initState = conf0 defState
      
      -- Detect creation
      maybeCreate (GLPictureCreateResp glcanvas) = Just (True,glcanvas)
      maybeCreate _ = Nothing
      
  in boxGUI $ proc (_,resp,conf) -> do
    -- State
    rec state <- iPre initState -< conf state
    
    -- Has the state changed?  If so, generate set request.
    -- stateChanged <- edgeBy maybeChanged initState -< state
    
    -- Creation
    (isCreated,glcanvas) <- hold (False,undefined) -< mapFilterE maybeCreate resp
    let doCreate = if isCreated then noEvent else Event ()
    
    -- Output.  Note that here we are again violating the
    -- rules of AFRP by sending an infinitely-dense stream
    -- of events, because we want a new picture at every
    -- sample time so that animation will work.
    let req = lMerge (tag doCreate (GLPictureCreateReq state))
                     (Event (GLPictureSetReq glcanvas state))
                     -- (tag stateChanged (GLPictureSetReq glcanvas state))
    returnA -< (req,())


-----------------------------------------------------------
-- Experimental Menu Support.

data MenuState = MenuState {
     msParentLabel :: String,
     msLabel :: String,
     msHelp  :: String,
     msEnabled :: Bool,
     msAddedCommand :: IO (),
     msAddedCommandWithDialog :: WX.Frame () -> IO (),
     msEnableDialog :: Bool
  } -- deriving (Eq, Show)

instance Eq MenuState where
    (==) (MenuState lpl ll lh lb lc lcd led)
         (MenuState rpl rl rh rb rc rcd red)
      = lpl == rpl && ll == rl && lh == rh && lb == rb
      && eqFunction lc rc && eqFunction lcd rcd && led == red

type MenuConf = MenuState -> MenuState

mtext :: String -> MenuConf
mtext m ms = ms {msLabel = m}

mtextWithDialog :: String -> MenuConf
mtextWithDialog m ms = ms {msLabel = m, msEnableDialog = True}

mhelp :: String -> MenuConf
mhelp m ms = ms {msHelp = m}

menabled :: Bool -> MenuConf
menabled e ms = ms {msEnabled = e}

mcommand :: IO () -> MenuConf
mcommand c ms = ms {msAddedCommand = c}

mcommandWithDialog  :: (WX.Frame () -> IO ()) -> MenuConf
mcommandWithDialog  c ms = ms {msAddedCommandWithDialog = c}

mcommandText :: IO () -> String -> MenuConf
mcommandText c t ms = ms {msAddedCommand = c, msLabel = t}

mcommandTextWithDialog :: (WX.Frame () -> IO ()) -> String -> MenuConf
mcommandTextWithDialog c t ms = ms {msAddedCommandWithDialog = c, msLabel = t}

menableDialog :: Bool -> MenuConf
menableDialog e ms = ms {msEnableDialog = e}


-- The menu GUI

menu :: MenuConf -> Box MenuConf (Event ())
menu conf0 =
  let -- Initial state
      defState = MenuState { msParentLabel = "&File", msLabel = "Default", msHelp ="", msEnabled = True
                           , msAddedCommand = return (), msAddedCommandWithDialog = \f -> return ()
                           , msEnableDialog = False}
      initState = conf0 defState
      
      -- Detect creation
      maybeCreate (MenuCreateResp b) = Just (True,b)
      maybeCreate _ = Nothing
      
      -- Detect button press
      maybeCommand MenuCommandEvent = Just ()
      maybeCommand _ = Nothing
  
  -- Make an ordinary, atomic signal function (in order to have access
  -- to the response/request mechanism), then lift it into a WXBox
  -- with wxBoxGUI.
  in boxGUI $ proc (_,resp,conf) -> do
    -- Keep track of the state.
    rec state <- iPre initState -< conf state
    
    -- Has the state changed?  If so, generate set request.
    stateChanged <- edgeBy maybeChanged initState -< state
    
    -- Ensure that the system hands us back a WX.Button ().
    (isCreated,menu) <- hold (False,undefined) -< mapFilterE maybeCreate resp
    let doCreate = if isCreated then noEvent else Event ()
    
    -- Merge create/set requests.
    let req = lMerge (tag doCreate (MenuCreateReq state))
                     (tag stateChanged (MenuSetReq menu state))
    
    -- Pass button presses through.
    let press = mapFilterE maybeCommand resp
    
    returnA -< (req,press)

{-
-----------------------------------------------------------
-- Experimental No Dialog Command Support.

data CommandState = CommandState {
     csEnabled :: Bool,
     csAddedCommand :: IO (),
     csAddedCommandWithDialog :: WX.Frame () -> IO (),
     csEnableDialog :: Bool
  } -- deriving (Eq, Show)

instance Eq CommandState where
    (==) (CommandState lb lc lcd led)
         (CommandState rb rc rcd red)
      = lb == rb && eqFunction lc rc && eqFunction lcd rcd && led == red

type CommandConf = CommandState -> CommandState

cenabled :: Bool -> CommandConf
cenabled e cs = cs {csEnabled = e}

ccommand :: IO () -> CommandConf
ccommand c cs = cs {csAddedCommand = c}

ccommandEnabled :: IO () -> Bool -> CommandConf
ccommandEnabled  c e cs = cs {csAddedCommand = c, csEnabled = e}

ccommandWithDialog  :: (WX.Frame () -> IO ()) -> CommandConf
ccommandWithDialog  c cs = cs {csAddedCommandWithDialog = c}

ccommandEnabledWithDialog  :: (WX.Frame () -> IO ()) -> Bool -> CommandConf
ccommandEnabledWithDialog  c e cs = cs {csAddedCommandWithDialog = c, csEnabled = e}

cenableDialog :: Bool -> CommandConf
cenableDialog e cs = cs {csEnableDialog = e}


-- The menu GUI

command :: CommandConf -> Box CommandConf ()
command conf0 =
  let -- Initial state
      defState = CommandState { csEnabled = True
                              , csAddedCommand = return (), csAddedCommandWithDialog = \f -> return ()
                              , csEnableDialog = False}
      initState = conf0 defState
      {-
      -- Detect creation
      maybeCreate (CommandCreateResp b) = Just (True,b)
      maybeCreate _ = Nothing
      -}
  -- Make an ordinary, atomic signal function (in order to have access
  -- to the response/request mechanism), then lift it into a WXBox
  -- with wxBoxGUI.
  in boxGUI $ proc (_,resp,conf) -> do
    -- Keep track of the state.
    rec state <- iPre initState -< conf state
    
    -- Has the state changed?  If so, generate set request.
    stateChanged <- edgeBy maybeChanged initState -< state
    {-
    -- Ensure that the system hands us back a WX.Button ().
    (isCreated,menu) <- hold (False,undefined) -< mapFilterE maybeCreate resp
    let doCreate = if isCreated then noEvent else Event ()
        -}
    let doCommand = if csEnabled state then noEvent else Event ()
    
    -- Merge create/set requests.
    let req = (tag doCommand (CommandSetReq state))
    
    returnA -< (req,())
-}

-----------------------------------------------------------
-- Mouse widget: this is problematic.  It just outputs the
-- current mouse position, relative to the client space of
-- the topmost frame.  There is currently no way to
-- translate this into coordinates relative to a particular
-- widget.

mouse :: Box () WX.Point
mouse = boxGUI $ proc (inp,_,_) -> do
  returnA -< (noEvent,mousePos inp)

{-
resizer :: Box () WX.Size
resizer = boxGUI $ proc (inp,_,_) -> do
  returnA -< (noEvent,windowSize inp)
-}

getEvent :: Box () (Event RawEvent)
getEvent = boxGUI $ proc (inp,_,_) -> do
  returnA -< (noEvent,rawEvent inp)

-----------------------------------------------------------
-- The WXFruit graphics library!  (Just wrappers for
-- wxHaskell functions, basically, adapting them to the
-- SOE/Haven graphics idiom.)  Just enough is implemented
-- here to get paddleball working.

-- Picture type: This corresponds to a wxHaskell repaint handler.
type Picture = [WX.Prop (WX.DC ())] -> WX.DC () -> WX.Rect -> IO ()

type GLPicture = WXCore.GLCanvas () -> WX.DC() -> WX.Rect -> [WX.Rect]-> IO ()

glBlank :: GLPicture
glBlank _ _ _ _ = return ()

-- Draw nothing.
blank :: Picture
blank _ _ _ = return () 

-- Fill the picture with a single color.
fill :: Picture
fill props dc varRect = picFill (rect varRect) props dc varRect

-- Draw an ellipse bounded by a rectangle.
ellipse :: WX.Rect -> Picture
ellipse rect props dc _ = WX.ellipse dc rect props

-- Draw a rectangle.
rect :: WX.Rect -> Picture
rect rect props dc _ = WX.drawRect dc rect props

-- Write graphical text.  Don't confuse this with wxtext;
-- it's an entirely different animal.
write :: String -> WX.Point -> Picture
write str pt props dc _ = WX.drawText dc str pt props

-- Make drawings use a particular color.
withColor :: WX.Color -> Picture -> Picture
withColor color pic props =
  pic (props ++ [WX.color WX.:= color, WX.brushColor WX.:= color])

-- Make rectangles and ellipses be filled in.
picFill :: Picture -> Picture
picFill pic props =
  pic (props ++ [WX.brushKind WX.:= WX.BrushSolid])

-- Combine two drawings.
picOver :: Picture -> Picture -> Picture
picOver pic1 pic2 props dc rect = do
  pic2 props dc rect
  pic1 props dc rect


-----------------------------------------------------------
-- Data types and functions for composition of widgets.
-- This is where the connection is made between the tree-
-- structure of WXGUIs and WXBoxes and the actual layout
-- of widgets sent to wxHaskell.  This is all rather
-- clumsy.

-- Contents widget tree: wxFruit version of WX.Layout.
data Contents = CEmpty
              | CLeaf Widget
              | CComp Orientation Contents Contents

data Widget = Button (WX.Button ())
            | Text (WX.StaticText ())
            | Picture (WX.Panel ())
            | GLPicture (WXCore.GLCanvas ())
            | Slider (WX.Slider ())
            -- Menu
            | MenuItem (WX.MenuItem ())

-- Compositional path to a widget.  A True entry means
-- go down the left/top branch of a WXContents tree; False
-- means go down the right/bottom.
type Path = [Bool]

-- Convert contents into layout.  The tricky part is
-- making sure that nested WXCComps with the same
-- orientation are merged into a single list (see
-- c2lList).
contents2layout :: Contents -> WX.Layout
contents2layout CEmpty = WX.empty
contents2layout (CLeaf (Button b)) = WX.widget b
contents2layout (CLeaf (Text t)) = WX.widget t
contents2layout (CLeaf (Picture p)) = WX.widget p
contents2layout (CLeaf (GLPicture g)) = WX.widget g
contents2layout (CLeaf (Slider s)) = WX.widget s
-- Menu
contents2layout (CLeaf (MenuItem b)) = WX.empty
contents2layout comp@(CComp orient c1 c2) =
  (if orient == Horiz then WX.row else WX.column) 5 (c2lList orient comp)

-- Merge similarly-orientated WXCComps together.
c2lList :: Orientation -> Contents -> [WX.Layout]
c2lList orient comp@(CComp orient' c1 c2) =
  if orient == orient'
    then (c2lList orient c1) ++ (c2lList orient c2)
    else [contents2layout comp]
c2lList _ CEmpty = []
c2lList _ c = [contents2layout c]

-- Destroy the windows in a contents tree.  This shouldn't
-- really be necessary, except wxHaskell doesn't currently
-- handle window destruction properly (IMO) -- simply
-- resetting the layout doesn't get rid of any windows
-- in the old layout.  This is really only needed if/when
-- widget-switching is implemented.
removeContents :: WX.Frame () -> Contents -> IO Bool
removeContents f CEmpty = return True
removeContents f (CLeaf (Button b)) = WXCore.windowDestroy b
removeContents f (CLeaf (Text t)) = WXCore.windowDestroy t
removeContents f (CLeaf (Slider s)) = WXCore.windowDestroy s
removeContents f (CLeaf (Picture p)) = WXCore.windowDestroy p
removeContents f (CLeaf (GLPicture g)) = WXCore.windowDestroy g
-- Menu
removeContents f (CLeaf (MenuItem m)) = WXCore.menuItemDelete m >> return True
removeContents f (CComp _ c0 c1) =
  do removeContents f c0
     removeContents f c1


-----------------------------------------------------------
-- The guts of the WXFruit system.  This is where the
-- connection between wxHaskell and Yampa is made.

-- State maintained across calls to react: the frame
-- window, the time of the last sample, and the contents
-- of the window (i.e. the entire widget tree).
-- type GUIState = (WX.Frame (),Int,Contents)
type GUIState = (WX.Frame (),Map.Map String (WX.Menu ()),Int,Contents)
type GUIRef = IORef GUIState

type RHandle = ReactHandle (RawInput,Event WidgetResp,()) (Event WidgetReq,())

-- Start a WXGUI running in a top-level frame.
startGUI :: String -> GUI () () -> IO ()
startGUI title (GUI g) = WX.start $ do
  -- Set up frame window and internal state.
  ch <- newChan
  f <- WX.frame [WX.text WX.:= title]
  defaultWindowsRawEvent f ch
  epoch <- getCurrentTime  
  gsr <- newIORef (f,Map.empty,epoch,CEmpty)
  
  -- Initialize Yampa.
  rh <- reactInit (initSense f) (actuate ch gsr) g
  
  -- Set up heartbeat timer to make periodic calls to react.
  -- This is necessary for animation.  Note that this is
  -- not the only place react is called from: button
  -- presses and other user actions may generate calls of
  -- their own.
  timer <- WX.timer f [WX.interval WX.:= 30, WX.on WX.command WX.:= respond ch gsr rh noEvent]
  
  return ()

defaultWindowsRawEvent w ch = do
    WXCore.windowOnMouse w False (\mouse -> writeChan ch $ Mouse mouse)
    -- WXCore.windowOnKeyChar w (\key -> writeChan ch $ Key key False)
    WXCore.windowOnKeyDown w (\key -> writeChan ch $ Key key False)
    WXCore.windowOnKeyUp w (\key -> writeChan ch $ Key key True)
    -- WXCore.windowOnIdle w (\idle -> writeChan ch $ Idle idle)

testChanel ch inp' = do
    emp <- isEmptyChan ch
    if emp
      then return inp'
      else do
          ev <- readChan ch
          let inp = inp' {rawEvent = Event ev}
          -- print $ inp
          return inp

-- Get an input sample from the OS.
getRawInput :: WX.Frame () -> IO RawInput
getRawInput f = do
  mouse <- WXCore.windowGetMousePosition f
  size <- WXCore.windowGetSize f
  -- defaultWindowsRawEvent f (\key -> print (show key))
  return RawInput {mousePos = mouse, windowSize = size, rawEvent = NoEvent}

-- The very first input sample.
initSense :: WX.Frame () -> IO (RawInput,Event WidgetResp,())
initSense f = do
  inp <- getRawInput f
  return (inp,Event Init,())

-- Push an input sample/widget response.
respond :: Chan RawEvent -> GUIRef -> RHandle -> Event WidgetResp -> IO ()
respond ch gsr rh resp = do
  -- Obtain input sample.
  (f,mu,prevt,c) <- readIORef gsr
  inp' <- getRawInput f
  inp <- testChanel ch inp'
  
  -- Make sure time's elapsed since the last call to react.
  -- With the timer set up in startGUI, this is probably
  -- unnecessary, but we'll leave it in to be safe.
  et <- getCurrentTime
  (dtf,t) <- ensureTimeElapses prevt et getCurrentTime
  writeIORef gsr (f,mu,t,c)
  
  -- Let Yampa work its magic.
  react rh (dtf,Just (inp,resp,()))
  return ()

-- Process an output sample (i.e. a widget request).
actuate :: Chan RawEvent -> GUIRef -> RHandle -> Bool -> (Event WidgetReq,()) -> IO Bool
actuate ch gsr rh _ (wre,_) = 
  do -- Handle requests, if any.
     (f,mu,t,prevc) <- readIORef gsr
     (resp,c,cch) <- handleWidgetReq ch gsr rh f [] prevc wre
     (f,mu,t,prevc) <- readIORef gsr
     
     -- Reset layout if contents changed.
     if cch
       then do WX.set f $ [WX.layout WX.:= WX.margin 5 (contents2layout c)]
                          ++ if Map.null mu then [] else [WX.menubar WX.:= elems mu]
               WX.windowReLayoutMinimal f
       else return ()
     writeIORef gsr (f,mu,t,c)
     
     -- Turn around and respond to the widgets, if necessary.
     -- Note that this causes a reentrant call to react.
     case resp of
       NoEvent -> return ()
       _ -> respond ch gsr rh resp
     
     return False

-- Handle a single widget request by making the appropriate
-- calls into wxHaskell.  Returns a response to the
-- request (if any), the new contents of the window, and
-- a boolean indicating whether the contents have changed.
handleWidgetReq :: Chan RawEvent -> GUIRef -> RHandle -> WX.Frame ()
                   -> Path -> Contents -> (Event WidgetReq)
                   -> IO (Event WidgetResp,Contents,Bool)

-- Do nothing if no request.
handleWidgetReq _ _ _ _ _ c NoEvent = return (noEvent,c,False)

-- Handle creation by calling the wxHaskell constructor with
-- properties corresponding to the given initial state,
-- removing any old contents, and returning the new widget.
handleWidgetReq ch gsr rh f path c (Event (ButtonCreateReq bstate)) =
  do let bprops = mkButtonProps ch gsr rh f path bstate
     b <- WX.smallButton f bprops
     removeContents f c
     return (Event (ButtonCreateResp b),CLeaf (Button b),True)
handleWidgetReq _ _ _ f path c (Event (TextCreateReq tstate)) =
  do let tprops = mkTextProps tstate
     t <- WX.staticText f tprops
     removeContents f c
     return (Event (TextCreateResp t),CLeaf (Text t),True)
handleWidgetReq _ _ _ f path c (Event (PictureCreateReq pstate)) =
  do let pprops = mkPictureProps pstate
     p <- WX.panel f pprops
     removeContents f c
     return (Event (PictureCreateResp p),CLeaf (Picture p),True)
handleWidgetReq ch _ _ f path c (Event (GLPictureCreateReq glpstate)) =
  do let pprops = mkGLPictureProps glpstate
         glsize = (glpsSize glpstate)
         glattrs = (glpsAttributes glpstate)
     gl <- (glpsCanvas glpstate) f glsize glattrs
     -- WX.set f [WX.on WX.paintRaw WX.:= (glpsPicture glpstate) gl]
     -- WX.set gl pprops
     defaultWindowsRawEvent gl ch
     removeContents f c
     return (Event (GLPictureCreateResp gl),CLeaf (GLPicture gl),True)
handleWidgetReq ch gsr rh f path c (Event (SliderCreateReq sstate)) =
  do -- Here we create the slider before setting up its properties
     -- because the event handler in the properties needs a pointer
     -- to the slider.
     s <- (if ssOrientation sstate == Horiz then WX.hslider else WX.vslider)
            f False (ssMin sstate) (ssMax sstate) []
     let sprops = mkSliderProps ch gsr rh path s sstate
     WX.set s sprops
     removeContents f c
     return (Event (SliderCreateResp s),CLeaf (Slider s),True)
-- Menu
handleWidgetReq ch gsr rh f path c (Event (MenuCreateReq mstate)) =
  do let mprops = mkMenuProps ch gsr rh f path mstate
     (f,pane,t,prevc) <- readIORef gsr
     file  <-
       let paneLabel = (msParentLabel mstate)
       in
         if Map.member paneLabel pane
           then  return $ pane Map.! paneLabel
           else  do
             newMenuPane <- WX.menuPane [WX.text WX.:= paneLabel]
             writeIORef gsr (f,Map.insert paneLabel newMenuPane pane,t,prevc)
             return newMenuPane
     m <- WX.menuItem file mprops
     removeContents f c
     return (Event (MenuCreateResp m),CLeaf (MenuItem m),True)

-- Handle creation requests by resetting the properties
-- of the widget.
handleWidgetReq ch gsr rh f path c (Event (ButtonSetReq b bstate)) =
  do let bprops = mkButtonProps ch gsr rh f path bstate
     WX.set b bprops
     return (noEvent,c,False)
handleWidgetReq _ _ _ f path c (Event (TextSetReq t tstate)) =
  do let tprops = mkTextProps tstate
     WX.set t tprops
     return (noEvent,c,False)
handleWidgetReq _ _ _ f path c (Event (PictureSetReq p pstate)) =
  do let pprops = mkPictureProps pstate
     WX.set p pprops
     WX.repaint p -- paint the new picture
     return (noEvent,c,False)
handleWidgetReq _ _ _ f path c (Event (GLPictureSetReq gl glpstate)) =
  do let pprops = mkGLPictureProps glpstate
     WX.set f [WX.on WX.paintRaw WX.:= dialogOrNothing glpstate glpsEnableDialog glpsPicture glpsPictureWithDialog f gl]
     WX.set gl pprops
     WX.repaint f -- paint the new picture
     return (noEvent,c,False)
  where
      dialogOrNothing state e elsecom thencom f gl =
        if (e state)
          then ((thencom state) f) gl
          else (elsecom state) gl
handleWidgetReq ch gsr rh f path c (Event (SliderSetReq s sstate)) =
  do -- Note that if properties other than the selection and
     -- enable are modified, this will not change them.
     let sprops = mkSliderProps ch gsr rh path s sstate
     WX.set s sprops
     return (noEvent,c,False)
-- Menu
handleWidgetReq ch gsr rh f path c (Event (MenuSetReq m mstate)) =
  do let mprops = mkMenuProps ch gsr rh f path mstate
     WX.set m mprops
     return (noEvent,c,False)
{-
-- Command
handleWidgetReq ch gsr rh f path c (Event (CommandSetReq cstate)) =
  do dialogOrNothing cstate csEnableDialog csAddedCommand csAddedCommandWithDialog f
     return (noEvent,c,False)
-}

-- Pass requests down the widget tree.
handleWidgetReq ch gsr rh f path c (Event (Comp orient (lreq,rreq))) =
  do (lresp,lc',lcch) <- handleWidgetReq ch gsr rh f (True:path) (subCont True c) lreq
     (rresp,rc',rcch) <- handleWidgetReq ch gsr rh f (False:path) (subCont False c) rreq
     -- If both children returned no response, don't compose
     -- them together.
     return (case (lresp,rresp) of
               (NoEvent,NoEvent) -> noEvent
               resp -> Event (DeComp resp),
             CComp orient lc' rc',lcch || rcch)
  where subCont which (CComp _ l r) =
          if which then l else r
        subCont _ _ = CEmpty -- contents tree has not been set up yet

-- Construct button properties list
mkButtonProps :: Chan RawEvent -> GUIRef -> RHandle -> (WX.Frame ()) -> Path -> ButtonState
                 -> [WX.Prop (WX.Button ())]
mkButtonProps ch gsr rh f path bs =
  [WX.text WX.:= bsLabel bs, WX.enabled WX.:= bsEnabled bs,
   WX.on WX.command WX.:= dialogOrNothing bs bsEnableDialog bsAddedCommand bsAddedCommandWithDialog f
     >> mkButtonCommand ch gsr rh path]

-- Button-press event handler.  Causes a call to react.
mkButtonCommand :: Chan RawEvent -> GUIRef -> RHandle -> Path -> IO ()
mkButtonCommand ch gsr rh path =
  do let e = pathify (Event ButtonCommandEvent) path
     respond ch gsr rh e

-- Contruct text label properties list.
mkTextProps :: TextState -> [WX.Prop (WX.StaticText ())]
mkTextProps ts =
  [WX.text WX.:= tsLabel ts]

-- Construct slider properties list.
mkSliderProps :: Chan RawEvent -> GUIRef -> RHandle -> Path
                 -> WX.Slider () -> SliderState -> [WX.Prop (WX.Slider ())]
mkSliderProps ch gsr rh path s ss =
  [WX.enabled WX.:= ssEnabled ss, WX.selection WX.:= ssSelection ss,
   WX.on WX.command WX.:= mkSliderCommand ch gsr rh path s]

-- Event handler called when slider is moved by user.
-- Causes a call to react.
mkSliderCommand :: Chan RawEvent -> GUIRef -> RHandle -> Path -> WX.Slider () -> IO ()
mkSliderCommand ch gsr rh path s =
  do sel <- WX.get s WX.selection
     let e = pathify (Event (SliderCommandEvent sel)) path
     respond ch gsr rh e

-- Construct picture widget properties list.
mkPictureProps :: PictureState -> [WX.Prop (WX.Panel ())]
mkPictureProps ps =
  [WX.clientSize WX.:= psSize ps, WX.on WX.paint WX.:= psPicture ps []]

mkGLPictureProps :: GLPictureState -> [WX.Prop (WXCore.GLCanvas ())]
mkGLPictureProps glps =
  [WX.clientSize WX.:= glpsSize glps]

-- Construct menu properties list
mkMenuProps :: Chan RawEvent -> GUIRef -> RHandle -> (WX.Frame ()) -> Path -> MenuState
               -> [WX.Prop (WX.MenuItem ())]
mkMenuProps ch gsr rh f path ms =
  [WX.text WX.:= msLabel ms, WX.enabled WX.:= msEnabled ms,
  -- WX.on WX.command WX.:= mkMenuCommand gsr rh path]
   WX.on WX.command WX.:= dialogOrNothing ms msEnableDialog msAddedCommand msAddedCommandWithDialog f >> mkMenuCommand ch gsr rh path]
  ++ let helpProp = msHelp ms in if helpProp == "" then [] else [WX.help WX.:= helpProp]

-- Menu-press event handler.  Causes a call to react.
mkMenuCommand :: Chan RawEvent -> GUIRef -> RHandle -> Path -> IO ()
mkMenuCommand ch gsr rh path =
  do let e = pathify (Event MenuCommandEvent) path
     respond ch gsr rh e

-- Construct a WXWidgetResp corresponding to a given
-- path.  This response will then be directed to the
-- correct widget (e.g. a button-pressed event will be sent
-- through the widget tree to the appropriate wxbutton).
pathify :: Event WidgetResp -> Path -> Event WidgetResp
pathify e [] = e
pathify e (True:path) = pathify (Event (DeComp (e,noEvent))) path
pathify e (False:path) = pathify (Event (DeComp (noEvent,e))) path


-----------------------------------------------------------
-- Utility functions copied from Yampa.

getCurrentTime :: IO Int
getCurrentTime =
  do (TOD sec psec) <- getClockTime
     let usec = fromIntegral psec / 1000000
     let msec = (fromIntegral ((sec - ctEpoch) * 1000)) + (usec / 1000)
     -- let dmsec = (fromIntegral $ msec) :: Double
     let imsec = round msec
     -- putStrLn ("getCurrentTime(): sec == " ++ show sec ++ ", usec == " ++
     --           show usec)
     -- putStrLn ("getCurrentTime() returning: " ++ (show imsec))
     return imsec
  where
    -- an arbitrary epoch, in seconds, to prevent Int wraparound:
    ctEpoch :: Integer
    ctEpoch = 1037286000


-- ensure an observable amount of time elapses by busy-waiting.
--
-- arguments:
--   t0 :: Int -- time of last sample
--   t1 :: Int -- current time, in milliseconds
--   getTime :: IO Int -- returns the current time, in milliseconds
-- result: (dtf,t1)
--   dtf -- the elapsed time since last sample, in seconds, as a Float
--   t1 -- the current time, in millisec.
--
-- We perform the floating point conversion, and perform our comparison
-- with respect to that conversion here to ensure that we are using
-- the same test as used by reactimate.
--
ensureTimeElapses :: Int -> Int -> IO Int -> IO (Double,Int)
ensureTimeElapses t0 t1 getTime = do
  let dt = t1 - t0
      dtf = (fromIntegral dt) / 1000
  if (dtf > 0) then return (dtf,t1)
               else do t' <- getTime
                       ensureTimeElapses t0 t' getTime

